from typing import Optional

from numpy import abs as _abs
from numpy import float64
from numpy import log as _log
from pandas import DataFrame, Series
from pandas import read_csv as _read_csv

from tompy.matrix import describe as _describe
from tompy.matrix import describe_meta as _describe_meta
from tompy.matrix import lr2crdd as _lr2crdd
from tompy.matrix import lr2sr as _lr2sr
from tompy.matrix import lr_metrics as _lr_metrics
from tompy.matrix import lr_sr_metrics as _lr_sr_metrics
from tompy.matrix import lr_sr_metrics_meta as _lr_sr_metrics_meta
from tompy.matrix import portfolio_price as _portfolio_price
from tompy.matrix import price2lr as _price2lr
from tompy.matrix import price2sr as _price2sr
from tompy.matrix import price_metrics as _price_metrics
from tompy.matrix import rolling_max as _rolling_max
from tompy.matrix import rolling_mean as _rolling_mean
from tompy.matrix import rolling_min as _rolling_min
from tompy.matrix import rolling_std as _rolling_std
from tompy.matrix import rolling_sum as _rolling_sum
from tompy.matrix import rolling_var as _rolling_var
from tompy.matrix import shift as _shift
from tompy.matrix import sr2lr as _sr2lr
from tompy.matrix import sr_metrics as _sr_metrics
from tompy.series import rolling_corr as _series_rolling_corr
from tompy.series import rolling_mean as _series_rolling_mean
from tompy.series import rolling_std as _series_rolling_std
from tompy.series import shift as _series_shift


def drop_weekends(df: DataFrame, /) -> DataFrame:
    return df[df.index.dayofweek < 5]


def read_csv(
    fpath: str, /, *, index_col: Optional[int] = 0, parse_dates: bool = True
) -> DataFrame:
    return _read_csv(fpath, index_col=index_col, parse_dates=parse_dates)


def write_csv(df: DataFrame, /, fpath: str) -> None:
    df.to_csv(fpath)


def describe(df: DataFrame, /) -> DataFrame:
    return DataFrame(
        data=_describe(df.values),
        index=_describe_meta().keys(),
        columns=df.columns,
        dtype=float64,
    )


def shift(df: DataFrame, /, lag: int) -> DataFrame:
    if lag == 0:
        return df
    return DataFrame(
        data=_shift(df.values, lag),
        index=df.index,
        columns=df.columns,
        dtype=float64,
    )


def rolling_sum(df: DataFrame, /, window: int) -> DataFrame:
    return DataFrame(
        data=_rolling_sum(df.values, window),
        index=df.index,
        columns=df.columns,
        dtype=float64,
    )


def rolling_mean(df: DataFrame, /, window: int) -> DataFrame:
    return DataFrame(
        data=_rolling_mean(df.values, window),
        index=df.index,
        columns=df.columns,
        dtype=float64,
    )


def rolling_std(df: DataFrame, /, window: int) -> DataFrame:
    return DataFrame(
        data=_rolling_std(df.values, window),
        index=df.index,
        columns=df.columns,
        dtype=float64,
    )


def rolling_var(df: DataFrame, /, window: int) -> DataFrame:
    return DataFrame(
        data=_rolling_var(df.values, window),
        index=df.index,
        columns=df.columns,
        dtype=float64,
    )


def rolling_min(df: DataFrame, /, window: int) -> DataFrame:
    return DataFrame(
        data=_rolling_min(df.values, window),
        index=df.index,
        columns=df.columns,
        dtype=float64,
    )


def rolling_max(df: DataFrame, /, window: int) -> DataFrame:
    return DataFrame(
        data=_rolling_max(df.values, window),
        index=df.index,
        columns=df.columns,
        dtype=float64,
    )


def rolling_corr(df: DataFrame, /, window: int, *, cidx: int = 0) -> DataFrame:
    s = df[df.columns[cidx]]
    df2 = DataFrame(index=df.index, columns=df.columns, dtype=float64)
    for col in df:
        df2[col] = _series_rolling_corr(s, df[col], window)
    return df2


def rolling_ema(df: DataFrame, /, window: int) -> DataFrame:
    return df.ewm(span=window, adjust=False).mean()


def price2sr(price: DataFrame, /) -> DataFrame:
    return DataFrame(
        data=_price2sr(price.values),
        index=price.index,
        columns=price.columns,
        dtype=float64,
    )


def price2lr(price: DataFrame, /) -> DataFrame:
    return DataFrame(
        data=_price2lr(price.values),
        index=price.index,
        columns=price.columns,
        dtype=float64,
    )


def price_metrics(
    price: DataFrame,
    /,
    ann_factor: float,
    *,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> DataFrame:
    return DataFrame(
        data=_price_metrics(
            price.values,
            ann_factor,
            sqrt_ann_factor=sqrt_ann_factor,
            mode=mode,
        ),
        index=_lr_sr_metrics_meta(mode).keys(),
        columns=price.columns,
        dtype=float64,
    )


def sr2lr(sr: DataFrame, /) -> DataFrame:
    return DataFrame(
        data=_sr2lr(sr.values),
        index=sr.index,
        columns=sr.columns,
        dtype=float64,
    )


def sr_metrics(
    sr: DataFrame,
    /,
    ann_factor: float,
    *,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> DataFrame:
    return DataFrame(
        data=_sr_metrics(
            sr.values, ann_factor, sqrt_ann_factor=sqrt_ann_factor, mode=mode
        ),
        index=_lr_sr_metrics_meta(mode).keys(),
        columns=sr.columns,
        dtype=float64,
    )


def lr2sr(lr: DataFrame, /) -> DataFrame:
    return DataFrame(
        data=_lr2sr(lr.values),
        index=lr.index,
        columns=lr.columns,
        dtype=float64,
    )


def lr2crdd(lr: DataFrame, /) -> dict[str, DataFrame]:
    lr0_cr, lr0_dd = _lr2crdd(lr.values)
    return {
        "LR_CR": DataFrame(
            data=lr0_cr, index=lr.index, columns=lr.columns, dtype=float64
        ),
        "LR_DD": DataFrame(
            data=lr0_dd, index=lr.index, columns=lr.columns, dtype=float64
        ),
    }


def lr_metrics(
    lr: DataFrame,
    /,
    ann_factor: float,
    *,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> DataFrame:
    return DataFrame(
        data=_lr_metrics(
            lr.values, ann_factor, sqrt_ann_factor=sqrt_ann_factor, mode=mode
        ),
        index=_lr_sr_metrics_meta(mode).keys(),
        columns=lr.columns,
        dtype=float64,
    )


def lr_sr_metrics(
    lr: DataFrame,
    sr: DataFrame,
    /,
    ann_factor: float,
    *,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> DataFrame:
    return DataFrame(
        data=_lr_sr_metrics(
            lr.values,
            sr.values,
            ann_factor,
            sqrt_ann_factor=sqrt_ann_factor,
            mode=mode,
        ),
        index=_lr_sr_metrics_meta(mode).keys(),
        columns=lr.columns,
        dtype=float64,
    )


def portfolio_price(
    price: DataFrame,
    weights: DataFrame,
    /,
    fee_rate: float,
) -> Series:
    price = price[weights.columns]
    price = price.loc[weights.index]
    return Series(
        data=_portfolio_price(price.values, weights.values, fee_rate),
        index=price.index,
        dtype=float64,
    )


def risk_std_lr_cc(price: DataFrame, /, window: int, *, col_c: str) -> Series:
    c = price[col_c]
    c1 = _series_shift(c, 1)
    lr = _log(c / c1)
    return _series_rolling_std(lr, window)


def risk_std_lr_oc(
    price: DataFrame, /, window: int, *, col_o: str, col_c: str
) -> Series:
    o = price[col_o]
    c = price[col_c]
    c1 = _series_shift(c, 1)
    lr = _log(o / c1)
    return _series_rolling_std(lr, window)


def risk_std_lr_co(
    price: DataFrame, /, window: int, *, col_o: str, col_c: str
) -> Series:
    o = price[col_o]
    c = price[col_c]
    lr = _log(c / o)
    return _series_rolling_std(lr, window)


def risk_std_sr_cc(price: DataFrame, /, window: int, *, col_c: str) -> Series:
    c = price[col_c]
    c1 = _series_shift(c, 1)
    sr = c / c1 - 1
    return _series_rolling_std(sr, window)


def risk_std_sr_oc(
    price: DataFrame, /, window: int, *, col_o: str, col_c: str
) -> Series:
    o = price[col_o]
    c = price[col_c]
    c1 = _series_shift(c, 1)
    sr = o / c1 - 1
    return _series_rolling_std(sr, window)


def risk_std_sr_co(
    price: DataFrame, /, window: int, *, col_o: str, col_c: str
) -> Series:
    o = price[col_o]
    c = price[col_c]
    sr = c / o - 1
    return _series_rolling_std(sr, window)


def risk_atr(
    price: DataFrame, /, window: int, *, col_h: str, col_l: str, col_c: str
) -> Series:
    h = price[col_h]
    l = price[col_l]
    c = price[col_c]
    c1 = _series_shift(c, 1)
    hl = h - l
    hc = _abs(h - c1)
    lc = _abs(l - c1)
    tr = DataFrame({"hl": hl, "hc": hc, "lc": lc}).max(axis=1, skipna=False)
    return _series_rolling_mean(tr, window)


def risk_ahl(
    price: DataFrame, /, window: int, *, col_h: str, col_l: str
) -> Series:
    h = price[col_h]
    l = price[col_l]
    hl = h - l
    return _series_rolling_mean(hl, window)


def risk_anr(
    price: DataFrame,
    /,
    window: int,
    *,
    col_o: str,
    col_h: str,
    col_l: str,
    col_c: str,
) -> Series:
    o = price[col_o]
    h = price[col_h]
    l = price[col_l]
    c = price[col_c]
    hl = h - l
    aco = _abs(c - o)
    noise = 1 - aco / hl
    return _series_rolling_mean(noise, window)
