from typing import Optional

from numpy import asarray as _asarray
from numpy import float64
from pandas import DataFrame, Series

from tompy.vector import describe as _describe
from tompy.vector import describe_meta as _describe_meta
from tompy.vector import lr2crdd as _lr2crdd
from tompy.vector import lr2sr as _lr2sr
from tompy.vector import lr_metrics as _lr_metrics
from tompy.vector import lr_sr_metrics as _lr_sr_metrics
from tompy.vector import lr_sr_metrics_meta as _lr_sr_metrics_meta
from tompy.vector import price2lr as _price2lr
from tompy.vector import price2sr as _price2sr
from tompy.vector import price_metrics as _price_metrics
from tompy.vector import price_terms_metrics as _price_terms_metrics
from tompy.vector import rolling_max as _rolling_max
from tompy.vector import rolling_mean as _rolling_mean
from tompy.vector import rolling_min as _rolling_min
from tompy.vector import rolling_std as _rolling_std
from tompy.vector import rolling_sum as _rolling_sum
from tompy.vector import rolling_var as _rolling_var
from tompy.vector import shift as _shift
from tompy.vector import sr2lr as _sr2lr
from tompy.vector import sr_metrics as _sr_metrics


def describe(s: Series, /) -> Series:
    return Series(
        data=_describe(s.values),
        index=_describe_meta().keys(),
        dtype=float64,
    )


def shift(s: Series, /, lag: int) -> Series:
    if lag == 0:
        return s
    return Series(data=_shift(s.values, lag), index=s.index, dtype=float64)


def rolling_sum(s: Series, /, window: int) -> Series:
    return Series(
        data=_rolling_sum(s.values, window),
        index=s.index,
        dtype=float64,
    )


def rolling_mean(s: Series, /, window: int) -> Series:
    return Series(
        data=_rolling_mean(s.values, window),
        index=s.index,
        dtype=float64,
    )


def rolling_std(s: Series, /, window: int) -> Series:
    return Series(
        data=_rolling_std(s.values, window),
        index=s.index,
        dtype=float64,
    )


def rolling_var(s: Series, /, window: int) -> Series:
    return Series(
        data=_rolling_var(s.values, window),
        index=s.index,
        dtype=float64,
    )


def rolling_min(s: Series, /, window: int) -> Series:
    return Series(
        data=_rolling_min(s.values, window),
        index=s.index,
        dtype=float64,
    )


def rolling_max(s: Series, /, window: int) -> Series:
    return Series(
        data=_rolling_max(s.values, window),
        index=s.index,
        dtype=float64,
    )


def rolling_corr(s1: Series, s2: Series, /, window: int) -> Series:
    return s1.rolling(window).corr(s2, ddof=1)


def rolling_ema(s: Series, /, window: int) -> Series:
    return s.ewm(span=window, adjust=False).mean()


def price2sr(price: Series, /) -> Series:
    return Series(
        data=_price2sr(price.values), index=price.index, dtype=float64
    )


def price2lr(price: Series, /) -> Series:
    return Series(
        data=_price2lr(price.values), index=price.index, dtype=float64
    )


def price_metrics(
    price: Series,
    /,
    ann_factor: float,
    *,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> Series:
    return Series(
        data=_price_metrics(
            price.values,
            ann_factor,
            sqrt_ann_factor=sqrt_ann_factor,
            mode=mode,
        ),
        index=_lr_sr_metrics_meta(mode).keys(),
        dtype=float64,
    )


def price_terms_metrics(
    price: Series,
    terms: dict[str, int],
    /,
    ann_factor: float,
    *,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> DataFrame:
    return DataFrame(
        data=_price_terms_metrics(
            price.values,
            list(terms.values()),
            ann_factor,
            sqrt_ann_factor=sqrt_ann_factor,
            mode=mode,
        ),
        index=_lr_sr_metrics_meta(mode).keys(),
        columns=terms.keys(),
        dtype=float64,
    )


def price_dterms_metrics(price: Series, /, *, mode: int = 0) -> DataFrame:
    n = price.shape[0]
    ann_factor = 252
    D = 1
    W = 5
    M = 21
    Y = M * 12
    terms = {
        "1D": 1 * D,
        "1W": 1 * W,
        "2W": 2 * W,
        "1M": 1 * M,
        "3M": 3 * M,
        "6M": 6 * M,
        "1Y": 1 * Y,
        "3Y": 3 * Y,
        "5Y": 5 * Y,
        "10Y": 10 * Y,
        "ITD": n,
    }
    return price_terms_metrics(price, terms, ann_factor, mode=mode)


def sr2lr(sr: Series, /) -> Series:
    return Series(data=_sr2lr(sr.values), index=sr.index, dtype=float64)


def sr_metrics(
    sr: Series,
    /,
    ann_factor: float,
    *,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> Series:
    return Series(
        data=_sr_metrics(
            sr.values, ann_factor, sqrt_ann_factor=sqrt_ann_factor, mode=mode
        ),
        index=_lr_sr_metrics_meta(mode).keys(),
        dtype=float64,
    )


def lr2sr(lr: Series, /) -> Series:
    return Series(data=_lr2sr(lr.values), index=lr.index, dtype=float64)


def lr2crdd(lr: Series, /) -> DataFrame:
    lr0_cr, lr0_dd = _lr2crdd(lr.values)
    return DataFrame(
        data=_asarray([lr0_cr, lr0_dd], dtype=float64).T,
        index=lr.index,
        columns=["LR_CR", "LR_DD"],
        dtype=float64,
    )


def lr_metrics(
    lr: Series,
    /,
    ann_factor: float,
    *,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> Series:
    return Series(
        data=_lr_metrics(
            lr.values, ann_factor, sqrt_ann_factor=sqrt_ann_factor, mode=mode
        ),
        index=_lr_sr_metrics_meta(mode).keys(),
        dtype=float64,
    )


def lr_sr_metrics(
    lr: Series,
    sr: Series,
    /,
    ann_factor: float,
    *,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> Series:
    return Series(
        data=_lr_sr_metrics(
            lr.values,
            sr.values,
            ann_factor,
            sqrt_ann_factor=sqrt_ann_factor,
            mode=mode,
        ),
        index=_lr_sr_metrics_meta(mode).keys(),
        dtype=float64,
    )
